import multiprocessing as mp
from lxml import etree
import collections
import re
import math
import os

class FreqList():
    def __init__(self, token_bound =r"[^A-Za-z0-9-]", min_length=1):
        self.token_bound_re = re.compile(token_bound, flags=re.U)
        self.word_count = collections.Counter()
        self.min_length = min_length

    def mp_from_file_list(self, file_list, pool_size=mp.cpu_count(), pipe=None):
        args = ((f, pipe) for f in file_list)
        jobs = []
        if pool_size > 4:
            pool_size = 4 # seems to perform better (vs 8) on my devices
        pool = mp.Pool(pool_size)
        num_files = len(file_list)
        job_counter = 0
        for job_wc in pool.imap_unordered(self.from_file, args):
            if job_wc:
                job_counter += 1
                self.word_count += job_wc
                per = math.floor(job_counter / num_files * 100)
                try:
                    pipe.send((job_counter, per))
                except BrokenPipeError as e:
                    pass
        try:
            pool.close()
        except Exception as e:
            raise(e)

    def from_file(self, args):
        pass

class XMLIterFreqList(FreqList):
    def __init__(self, token_bound =r"[^A-Za-z0-9'-]+", elem_list=[], att_dict={}, 
                 att_logic="AND", min_length=1):
        super().__init__(token_bound)
        self.elem_list = elem_list
        self.att_dict = att_dict
        self.att_logic = att_logic


class XMLFreqList(FreqList):
    def __init__(self, token_bound =r"[^A-Za-z0-9'-]", elem_list=[], att_dict={}, 
                 att_logic="AND", min_length=1):
        super().__init__(token_bound, min_length)
        self.elem_list = elem_list
        self.att_dict = att_dict
        self.att_logic = att_logic

    def from_file(self, args):
        f, pipe = args
        fn = os.path.basename(f)
        tree = etree.parse(f)
        wc = collections.Counter()
        if self.elem_list:
            for elem in self.elem_list:
                elem_str = ".//{}".format(elem)
            for node in tree.findall(elem_str):
                wc += self.count_in_node(node)
        return wc

    def count_in_node(self, node):
        wc = collections.Counter()
        use = True
        if self.att_dict:
            node_att_dict = {}
            try:            
                if self.att_logic == "AND":  # all attributes have to be in specified value(s)
                    for k in self.att_dict.keys():
                        node_k = node.get(k)
                        if node_k not in self.att_dict[k]:
                            use = False
                            break
                elif self.att_logic == "OR":
                    use = False
                    for k in self.att_dict.keys():
                        node_k = node.get(k)
                        if node_k in self.att_dict[k]:
                            use = True
                            break
            except Exception as e:
                raise(e)
        if use and node.text:
            text = " ".join(node.itertext())
            words = self.token_bound_re.split(text)
            words = [w for w in words if len(w)>=self.min_length]
            wc = collections.Counter(words)
        return wc


class Concordance():
    def __init__(self, search_str, is_regex=True, len_left=60, len_right=60, kwic_unit="char", 
                 add_sentence=False, sent_delim_re=re.compile(r"[\.\?!]_\."), token_bound =r"[^A-Za-z0-9-]"):
        self.search_str = search_str
        if type(self.search_str) == "str"  : # if not already compiled re
            if self.is_regex:
                self.search_re = re.compile(search_str)
            else:
                self.search_re = re.escape(search_str)
        else:
            self.search_re = search_str
        self.conc_lines = []
        self.lines = []
        self.len_left = len_left
        self.len_right = len_right
        self.kwic_unit = "char"
        self.add_sentence = add_sentence
        self.sent_delim_re = sent_delim_re
        self.token_bound_re = re.compile(token_bound, flags=re.U)
 

    def mp_from_file_list(self, file_list, is_plaintext=True, at_once=True, elem_list=[], 
                          attr_as_meta=True, pipe=None):
        args = ((f, is_plaintext, at_once, elem_list, attr_as_meta, pipe) for f in file_list)
        jobs = []
        pool_size = mp.cpu_count()
        last_hits = 0
        if pool_size > 4:
            pool_size = 4 # seems to perform better (vs 8) on my devices
        pool = mp.Pool(pool_size)
        num_files = len(file_list)
        job_counter = 0
        for job_lines in pool.imap_unordered(self.from_file, args):
            if job_lines:
                job_counter += 1
                self.conc_lines += job_lines
                hits = len(self.conc_lines)
                per = math.floor(job_counter / num_files * 100)
                #print(hits, job_counter, per)    
                try:
                    pipe.send((hits, job_counter, per))
                except BrokenPipeError as e:
                    pass
        pool.close()
        pool.terminate()
        
    def from_file(self, args):
        f, is_plaintext, at_once, elem_list, attr_as_meta, pipe = args
        fn = os.path.basename(f)
        if is_plaintext:
            with open(f, "r") as h:
                text = h.read()
                file_df = self.lines_from_text(text, fn)
        else:
            if elem_list: # only specific text nodes
                if at_once:
                    lines = self.parse_xml_file(f, attr_as_meta, fn)
                else:
                    context = etree.iterparse(f, events=('end',), tag=elem_list)
                    file_df = self.fast_conc_iter(context, attr_as_meta, fn)
            else: # all text nodes
                if at_once:
                    lines = self.parse_xml_file(f, attr_as_meta, fn)
                else:
                    context = etree.iterparse(f, events=('end',))
                    file_df = self.fast_conc_iter(context, attr_as_meta, fn)
        return lines

    def from_file_list(self, file_list, is_plaintext=True, at_once=True, elem_list=[], attr_as_meta=True,
                       pipe=None):
        self.elem_list = elem_list
        t = len(file_list)
        for i, f in enumerate(file_list):
            fn = os.path.basename(f)
            print(i+1, "/", t)
            if is_plaintext:
                with open(f, "r") as h:
                    text = h.read()
                    file_df = self.lines_from_text(text)
            else:
                if elem_list: # only specific text nodes
                    context = etree.iterparse(f, events=('end',), tag=elem_list)
                else: # all text nodes
                    context = etree.iterparse(f, events=('end',))
                file_df = self.fast_conc_iter(context, attr_as_meta)
            hits = len(file_df.index)
            if hits > 0:
                try:
                    pipe.send(hits)
                except Exception as e:
                    pass
            self.conc_lines += file_ines
 
    def lines_from_node(self, elem, attr_as_meta=False, fn=None):
        lines = []
        if elem.text:
            text = " ".join(elem.itertext())
            text = text.replace("\n", " ").replace("  ", " ").strip()
            for match in self.search_re.finditer(text):
                metadata = {} # from attributes, if asst_as_meta
                if attr_as_meta:
                    attr = elem.attrib
                    metadata = {k:v for k,v in attr.items()}
                l = self.kwic_from_match(text, match, metadata)
                if fn:
                    l["filename"] = fn
                if metadata:
                    l.update(metadata)
                lines.append(l)
        return lines

    def lines_from_text(self, text, fn=None):
        lines = []
        text = fix_POS(text)
        for match in self.search_re.finditer(text):
                l = self.kwic_from_match(text, match, metadata)
                if fn:
                    l['filename'] = fn
                lines.append(l)
        lines_df = pd.DataFrame().from_dict(lines)
        return lines_df

    def kwic_from_match(self, text, m, metadata={}):
        kwic = {}
        kwic['key'] = m.group(0).replace("\n", " ")
        if self.kwic_unit != "token":
            l_start = m.start()-self.len_left
            if l_start < 0:
                l_start = 0
            kwic['left'] = text[l_start:m.start()].replace("\n", " ").strip()
            kwic['right'] = text[m.end():m.end()+self.len_right].replace("\n", " ").strip()
        else:
            tmp_l = text[:m.start()]  # all left
            tmp_r = text[m.send():] # all right
            words_l = [w for w in self.token_bound_re.split(tmp_l) if not w.startswith("_")]  # exlcuding pos tags
            words_r = [w for w in self.token_bound_re.split(tmp_r) if not w.startswith("_")] 
            if words_r > self.len_r:
                words_r = words_r[:-self.len_r]  # only the first len_r words
            if words_l > self.len_l:
                words_l = words_l[-self.len_l:]
            kwic['left'] = " ".join(words_l)
            kwic['right'] = " ".join(words_r)
        if self.add_sentence:
            kwic['sentence'] = self.extract_sentence_from_match(text, m)
        if metadata:
            pass
        return kwic

    def extract_sentence_from_match(self, text, m):
        try:
            sent_end = m.end() + self.sent_delim_re.search(text[m.end():]).end()
        except:
            sent_end = len(text)
        sent_start = 0
        for pm in self.sent_delim_re.finditer(text[:m.start()]):
            sent_start = pm.end()
        sent = '"{}"'.format(text[sent_start:sent_end].strip().replace("\n", " "))
        return sent

    def parse_xml_file(self, f, attr_as_meta=False, fn=None):
        lines = []
        tree = etree.parse(f)
        for elem in tree.findall(".//u"):
            l = self.lines_from_node(elem, attr_as_meta, fn)
            if l:
                lines += l
        return lines

    def fast_conc_iter(self, context, attr_as_meta=False, fn=None):
        """
        fast_iter is useful if you need to free memory while iterating through a
        very large XML file.

        http://lxml.de/parsing.html#modifying-the-tree
        Based on Liza Daly's fast_iter
        http://www.ibm.com/developerworks/xml/library/x-hiperfparse/
        See also http://effbot.org/zone/element-iterparse.htm
        
        Slightly customized for the concordancer (e.g. return values, fixed args)

        """
        lines = []
        for event, elem in context:
            lines += self.lines_from_node(elem, attr_as_meta, fn)
            # It's safe to call clear() here because no descendants will be
            # accessed
            elem.clear()
            # Also eliminate now-empty references from the root node to elem
            for ancestor in elem.xpath('ancestor-or-self::*'):
                while ancestor.getprevious() is not None:
                    del ancestor.getparent()[0]
        del context
        return lines


class XMLConcordance(Concordance):
    def __init__(search_str, elem_list=[], att_values={}, att_logic="AND", att_as_meta=True, is_regex=True,
                 len_left=60, len_right=60, kwic_unit="char", 
                 add_sentence=False, sent_delim_re=None):
        self.elem_list = elem_list
        self.att_values = att_values
        self.att_logic = att_logic


class TXTConcordance(Concordance):
    def __init__(search_str, is_regex=True, len_left=60, 
                 len_right=60, kwic_unit="char", 
                 add_sentence=False, sent_delim_re=None):
        pass
