import multiprocessing as mp
import pandas as pd
import tkinter as tk
import tkinter.font as tkFont
import tkinter.ttk as ttk
from tkinter import filedialog
import xlsxwriter
import datetime
import os
import sys
import re
import obc2conc
import random
import configparser


class OBCInterface:
    def __init__(self):
        self.root = tk.Tk()
        try:
            self.root.tk.call('console', 'hide')
        except tk.TclError:
            pass
        self.base_dir = os.path.dirname(os.path.realpath(sys.argv[0]))
        self.user_dir = os.path.join(self.base_dir, "user")
        self.conc_dir = os.path.join(self.user_dir, "concordances")
        self.sub_dir = os.path.join(self.user_dir, "subcorpora")
        try:
            os.makedirs(self.conc_dir)
        except IOError:
            pass
        try:
            os.makedirs(self.sub_dir)
        except IOError:
            pass

        self.draw_ui()
        self.maximize()
        self.root.mainloop()

    def maximize(self):
        toplevel = self.root.winfo_toplevel()
        try:
            toplevel.wm_state('zoomed')
        except tk.TclError:
            w = self.root.winfo_screenwidth()
            h = self.root.winfo_screenheight() - 60
            geom_string = "%dx%d+0+0" % (w,h)
            toplevel.wm_geometry(geom_string)

    def draw_ui(self):
        frame_style = ttk.Style()
        frame_style.configure('obcFrame.TFrame', background='deepskyblue')
        self.root.title("OBC2Conc")
        self.root.columnconfigure(0, weight=1) 
        self.root.rowconfigure(0, weight=1)
        self.main_frame = ttk.Frame(self.root, padding="3 3 3 3", style="obcFrame.TFrame")
        self.main_frame.grid(row=0, column=0, sticky="news")
        self.main_frame.columnconfigure(0, weight=1)
        self.main_frame.rowconfigure(0, weight=1)
        self.main_book = ttk.Notebook(self.main_frame)
        self.main_book.grid(row=0, column=0, sticky="news")
        self.conc_frame = ConcordanceFrame(root=self.root, main=self)
        self.main_book.add(self.conc_frame, text="Concordance")
        self.freq_frame = FreqFrame(root=self.root, main=self)
        self.main_book.add(self.freq_frame, text="Token list")
        self.conf_frame = ConfigFrame(root=self.root, main=self)
        self.main_book.add(self.conf_frame, text="Settings")


class InfoFrame(tk.Frame):
    def __init__(self, root, main, *args, **kwargs):
        tk.Frame.__init__(self, *args, **kwargs)
        self.main = main
        self.root = root
        self.draw_ui()

    def draw_ui(self):
        self.columnconfigure(0, weight=0)
        self.rowconfigure(0, weight=0)
        
class ConfigFrame(tk.Frame):

    def __init__(self, root, main, *args, **kwargs):
        tk.Frame.__init__(self, *args, **kwargs)
        self.main = main
        self.root = root
        self.base_dir = os.path.dirname(os.path.realpath(sys.argv[0]))
        self.user_dir = os.path.join(self.base_dir, "user")
        self.config_path = os.path.join(self.user_dir, "config.ini")
        self.draw_ui()

    def draw_ui(self):
        self.columnconfigure(0, weight=0)
        self.rowconfigure(0, weight=0)
        self.auto_open = tk.IntVar()
        self.add_search_info = tk.IntVar()
        self.csv_quotechar = tk.StringVar()
        self.csv_delimiter = tk.StringVar()
        self.auto_open_check = ttk.Checkbutton(self, text="Open files after export",
                                               variable=self.auto_open, onvalue=1, offvalue=0)
        self.auto_open_check.grid(column=0, row=1, columnspan=2, sticky="w")
        self.config_button = ttk.Button(self, text="Save configuration as default",
                                        command=self.on_config_click)
        self.config_button.grid(row=5, column=0, columnspan=2, sticky="ew")
        self.check_config()
        self.auto_open.set(self.config.get("Export","AutoOpen"))
        for child in self.winfo_children():
            child.grid_configure(padx=3, pady=3)

    def on_config_click(self):
            self.config.set("Export","AutoOpen", str(self.auto_open.get()))
            
    def create_default_config(self):
        self.config = configparser.SafeConfigParser()
        self.config.optionxform = str
        self.config.add_section("Export")
        self.config.set("Export","AutoOpen","1")
        with open(self.config_path, 'w') as f:
            self.config.write(f)


    def check_config(self):
        self.config = configparser.SafeConfigParser()
        if not os.path.exists(self.config_path):
            self.create_default_config()
        else: # do basic sanity checks on config, set to defaults if values off
            self.config.optionxform = str
            self.config.read(self.config_path)
            if self.config.get("Export","AutoOpen") not in ("0","1"):
                self.config.set("Export","AutoOpen","1")


class FreqFrame(tk.Frame):
    def __init__(self, root, main, *args, **kwargs):
            tk.Frame.__init__(self, *args, **kwargs)
            self.base_dir = os.path.dirname(os.path.realpath(sys.argv[0]))
            self.user_dir = os.path.join(self.base_dir, "user")
            self.list_dir = os.path.join(self.user_dir, "lists")
            self.root = root
            self.main = main
            self.draw_ui()

    def draw_ui(self):
        self.start_year = tk.IntVar()
        self.end_year = tk.IntVar()
        self.columnconfigure(0, weight=1)
        self.rowconfigure(0, weight=1)
        self.rowconfigure(1, weight=0)
        self.rowconfigure(2, weight=0)
        self.info_label = ttk.Label(self, text="", wraplength=600)
        self.info_label.grid(column=0, row=2, sticky="new", columnspan=5)
        self.tool_frame = ttk.Frame(self, padding="5 5 5 5")
        self.tool_frame.grid(row=1, column=0, sticky="news")
        self.display_frame = ttk.Frame(self, padding="5 5 5 5")
        self.display_frame.rowconfigure(0, weight=1)
        self.display_frame.rowconfigure(1, weight=0)
        self.display_frame.columnconfigure(0, weight=1)
        self.display_frame.columnconfigure(1, weight=0)
        self.display_frame.grid(row=0, column=0, sticky="news")
        self.stylistics_frame = ttk.Frame(self, padding="5 5 5 5")
        self.stylistics_frame.grid(row=0, column=1, sticky="news")
        #  display frame
        self.list_header = ["#", "token", "frequency"]  #, "spread"]
        self.list_tree = ttk.Treeview(self.display_frame, columns=self.list_header, show="headings")
        self.ybar = ttk.Scrollbar(self.display_frame, orient="vertical",
                                  command=self.list_tree.yview)
        self.xbar = ttk.Scrollbar(self.display_frame, orient="horizontal",
                                  command=self.list_tree.xview)
        self.list_tree.configure(yscrollcommand=self.ybar.set, xscrollcommand=self.xbar.set)
        for col in self.list_header:
            self.list_tree.heading(col, text=col.upper(),
                                   command=lambda c=col: self.sort_list_by(self.list_tree, c, 0))

        self.list_tree.column("#", width=tkFont.Font().measure("1000000"))
        self.list_tree.column("token", width=tkFont.Font().measure("antidisestablishmentarianism_NN1"))
        self.list_tree.column("frequency", width=tkFont.Font().measure("1000000"))

        self.ybar.grid(row=0, column=1, sticky="news")
        self.xbar.grid(row=1, column=0, sticky="news")
        self.list_tree.grid(row=0, column=0, sticky='news')
        #  tool frame
        tk.Label(self.tool_frame, text="Sex").grid(row=1,column=0, sticky="nw")
        self.sex_list = tk.Listbox(self.tool_frame, exportselection=0,
                                   selectmode="multiple", height=3, width=7)
        self.sex_list.grid(row=1, column=1, sticky="nw")
        tk.Label(self.tool_frame, text="Role").grid(row=1, column=2, sticky="nw")
        self.role_list = tk.Listbox(self.tool_frame, exportselection=0,
                                   selectmode="multiple", height=6, width=10)
        self.role_list.grid(row=1, column=3, sticky="nw")
        tk.Label(self.tool_frame, text="Class").grid(row=1, column=4, sticky="nw")
        self.class_list = tk.Listbox(self.tool_frame, exportselection=0, selectmode="multiple", 
                                      height=3, width=9)
        self.class_list.grid(row=1, column=5, sticky="nw")
        tk.Label(self.tool_frame, text="HISCLASS").grid(column=6, row=1, sticky="nw")
        self.hisclass_list = tk.Listbox(self.tool_frame, exportselection=0,
                                   selectmode="multiple", height=14, width=40)
        self.hisclass_list.grid(row=1, column=7, sticky="nw", columnspan=2)

        tk.Label(self.tool_frame, text="From year").grid(column=9, row=1, sticky="nw")
        tk.Label(self.tool_frame, text="to year").grid(column=11, row=1, sticky="nw")
        self.start_year_entry = ttk.Entry(self.tool_frame, width=4, textvariable=self.start_year)
        self.end_year_entry = ttk.Entry(self.tool_frame, width=4, textvariable=self.end_year)
        self.start_year_entry.grid(row=1, column=10, sticky="nw")
        self.end_year_entry.grid(row=1, column=12, sticky="nw")

        self.start_button = ttk.Button(self.tool_frame, text="Start",
                                        command=self.on_start_click)
        self.start_button.grid(row=2, column=0, columnspan=7, sticky="ew")



        self.export_csv_button = ttk.Button(self.tool_frame, text="Export to .csv",
                                            command=self.on_export_csv_click)
        self.export_csv_button.grid(column=7, row=2, columnspan=1, sticky="ew")
        self.export_csv_button["state"] = "disabled"
        self.export_xlsx_button = ttk.Button(self.tool_frame, text="Export to .xlsx", command=self.on_export_xlsx_click)
        self.export_xlsx_button.grid(column=8, row=2, columnspan=1, sticky="ew")
        self.export_xlsx_button["state"] = "disabled"
        self.export_html_button = ttk.Button(self.tool_frame, text="Export to .html", command=self.on_export_html_click)
        self.export_html_button.grid(column=9, row=2, columnspan=2, sticky="ew")
        self.export_html_button["state"] = "disabled"
        self.export_tex_button = ttk.Button(self.tool_frame, text="Export to .tex", command=self.on_export_tex_click)
        self.export_tex_button.grid(column=11, row=2, columnspan=2, sticky="ew")
        self.export_tex_button["state"] = "disabled"
        
        self.bind('<Return>', self.on_start_click)


        for s in ["female", "male", "unknown"]:
            self.sex_list.insert("end", s)
        for r in ["defendant", "judge", "lawyer", "victim", "witness", "unknown"]:
            self.role_list.insert("end", r)
        for c in ["higher (1-5)", "lower (6-13)", "unknown"]:
            self.class_list.insert("end", c)
        for h in ['1 - Higher managers','2 - Higher professionals','3 - Lower managers',
                  '4 - Lower professionals, clerical and sales personnel',
                  '5 - Lower clerical and sales personnel','6 - Foremen', 
                  '7 - Foremen and skilled workers', '8 - Farmers and fishermen', 
                  '9 - Lower-skilled workers', '10 - Lower-skilled farm workers',
                  '11 - Unskilled workers', '12 - Unskilled farm workers',  
                  '13 - Unspecified workers', 'unknown']:
            self.hisclass_list.insert("end", h)
        self.start_year.set("1720")
        self.end_year.set("1913")

        for child in self.tool_frame.winfo_children():
            child.grid_configure(padx=5, pady=5)


            
    def on_export_csv_click(self):
        col_order = ['token', 'frequency']
        if len(self.list_df.index)> 0:
            options = {}
            options['defaultextension'] = ".csv"
            options['filetypes'] = [('CSV files', '.csv')]
            options['initialdir'] = self.list_dir
            options['initialfile'] = 'frequency_list.csv'
            options['parent'] = self
            options['title'] = 'Save'
            file_str = filedialog.asksaveasfilename(**options)
            if file_str:
                self.list_df.to_csv(file_str, columns=col_order)
                auto = bool(self.main.conf_frame.auto_open.get())    
                if auto:
                    open_file_externally(file_str)
            

    def on_export_html_click(self):
        pd.set_option('display.max_colwidth', -1)
        col_order = ['token', 'frequency']
        if len(self.list_df.index)> 0:
            options = {}
            options['defaultextension'] = ".html"
            options['filetypes'] = [('HTML files', '.html')]
            options['initialdir'] = self.list_dir
            options['initialfile'] = 'frequency_list.html'
            options['parent'] = self
            options['title'] = 'Save'
            file_str = filedialog.asksaveasfilename(**options)
            if file_str:
                html = self.list_df.to_html(columns=col_order)
                with open(file_str, "w") as h:
                    h.write(html)
                auto = bool(self.main.conf_frame.auto_open.get())    
                if auto:
                    obc2conc.open_file_externally(file_str)

    def on_export_tex_click(self):
        col_order = ['token', 'frequency']
        if len(self.list_df.index)> 0:
            options = {}
            options['defaultextension'] = ".tex"
            options['filetypes'] = [('LaTeX files', '.tex')]
            options['initialdir'] = self.list_dir
            options['initialfile'] = 'frequency_list.tex'
            options['parent'] = self
            options['title'] = 'Save'
            file_str = filedialog.asksaveasfilename(**options)
            if file_str:
                tex = self.list_df.to_latex(columns=col_order)
                with open(file_str, "w") as h:
                    h.write(tex)


    def on_export_xlsx_click(self):
        col_order = ['token', 'frequency']
        if len(self.list_df.index > 0):
            options = {}
            options['defaultextension'] = ".xlsx"
            options['filetypes'] = [('Excel files', '.xlsx')]
            options['initialdir'] = self.list_dir
            options['initialfile'] = 'frequency_list.xlsx'
            options['parent'] = self
            options['title'] = 'Save'
            file_str = filedialog.asksaveasfilename(**options)
            if file_str:
                writer = pd.ExcelWriter(file_str, engine='xlsxwriter')
                workbook = writer.book
                self.list_df.to_excel(writer, sheet_name="Frequency list", columns=col_order)      
                writer.save()
                auto = bool(self.main.conf_frame.auto_open.get())    
                if auto:
                    obc2conc.open_file_externally(file_str)


    def add_options_from_lists(self, options):
        sex = [self.sex_list.get(i) for i in self.sex_list.curselection()]
        if len(sex) in (1, 2):  # if none or all 3 are selected, no filtering needed
            sex = [s[0] if s!="unknown" else "" for s in sex]  # to get "f", "m", ""
            options['att_dict']['sex'] = sex
        role = [self.role_list.get(i) for i in self.role_list.curselection()]
        if len(role) not in (0, 6):  # if none or all 6 are selected, no filtering needed
            role = [r.title() if r!="unknown" else "" for r in role]
            options['att_dict']['role']= role
        binclass = [self.class_list.get(i) for i in self.class_list.curselection()]
        if len(binclass) in (1, 2): 
            binclass = [c if c!="unknown" else "" for c in binclass]
            options['att_dict']['class'] = binclass
        hisclass = [self.hisclass_list.get(i) for i in self.hisclass_list.curselection()]
        if len(hisclass) not in (0, 14): 
            hisclass = [c if c!="unknown" else "" for c in hisclass]
            options['att_dict']['hisclass'] = hisclass
        return options

    def on_start_click(self):
        if self.start_button["text"] == "Start":
            self.cancelled = False
            self.clear_list_view()
            self.export_csv_button["state"] = "disabled"
            self.export_xlsx_button["state"] = "disabled"
            self.export_tex_button["state"] = "disabled"
            self.export_html_button["state"] = "disabled"
            self.start_button.configure(text="Cancel", state="normal")
            self.info_label.config(text="Creating list. Please wait.")
            self.root.update_idletasks()
            self.list_start_time = datetime.datetime.now() 
            options = {}
            options['att_dict'] = {}
            options = self.add_options_from_lists(options)
            options['start_year'] = self.start_year.get()
            options['end_year'] = self.end_year.get()
            options['iterate'] = False
            options['att_logic'] = "AND"
            manager = mp.Manager()
            self.pipe, worker_pipe = mp.Pipe(duplex=False)
            self.return_shared = manager.list()
            self.list_job = mp.Process(target=obc2conc.worker_create_list,
                                       args=(options, self.return_shared, worker_pipe,))
            self.list_job.start()
            self.root.update_idletasks()
            self.root.after(1000, self.check_list_status)
        else:
            self.start_button["text"] = "Start"
            self.cancelled = True
            self.info_label["text"] = "Cancelled."
            self.pipe.close()
            self.list_job.terminate()
            self.root.after_cancel(self.after_job)
            self.root.update_idletasks()
            

    def check_list_status(self):
        try:
            if self.list_job.is_alive():
                try:
                    if self.pipe.poll(timeout=1):
                        num_files, per = self.pipe.recv()
                        self.info_label["text"] = "Creating list. Processed >{} files so far (~{}% done)".format(num_files, per)
                        self.root.update_idletasks()
                except BrokenPipeError as e:
                    pass
                try:
                    self.after_job = self.root.after(100, self.check_list_status)
                except AttributeError:
                    pass
            else:
                try:
                    self.root.after_cancel(self.after_job)
                except AttributeError:
                    pass # e.g. very quick runs
                finally:
                    self.start_button["text"] = "Start"
                    self.pipe.close()
                    if self.cancelled:
                        self.clear_list_view()
                    else:
                        self.word_count = self.return_shared[0]
                        self.finish_list()
        except AttributeError as e:
            raise(e)


    def finish_list(self):
        self.list_end_time = datetime.datetime.now() 
        diff_in_min = (self.list_end_time-self.list_start_time)/ datetime.timedelta(seconds=1)
        self.num_tokens = sum(self.word_count.values())
        self.num_types = len(self.word_count.most_common())
        self.info_label["text"] = "Found {:,} tokens and {:,} types. (List generation took {} seconds).".format(self.num_tokens, self.num_types, diff_in_min)
        self.export_csv_button["state"] = "normal"
        self.export_xlsx_button["state"] = "normal"
        self.export_tex_button["state"] = "normal"
        self.export_html_button["state"] = "normal"
        self.display_list_view()

    def display_list_view(self):
        self.list_df = pd.DataFrame.from_records(self.word_count.most_common(), columns=["token", "frequency"])
        self.clear_list_view()
        update_every = 200
        if self.num_tokens > 0:
            ind = 1
            for tok, freq in self.word_count.most_common():
                self.add_list_line(ind, tok, freq)
                ind += 1
                if ind % update_every == 0:
                    self.root.update_idletasks()

    def add_list_line(self, ind, tok, freq):
        row = {"#":ind, "token": tok, "frequency": freq}
        columns = [row[k] for k in self.list_header]
        self.list_tree.insert('', 'end', values=columns)


    def clear_list_view(self):
        self.list_tree.delete(*self.list_tree.get_children())



    def sort_list_by(self, tree, col, descending):
        if col in ["#", "frequency"]:
            data = [(int(tree.set(child, col)), child) \
                for child in tree.get_children('')]
        else:
            data = [(tree.set(child, col), child) \
                for child in tree.get_children('')]
        
        data.sort(reverse=descending)
        for ix, item in enumerate(data):
            tree.move(item[1], '', ix)
        tree.heading(col, command=lambda col=col: self.sort_list_by(tree, col, int(not descending)))



class ConcordanceFrame(tk.Frame):

    def __init__(self, root, main, *args, **kwargs):
        tk.Frame.__init__(self, *args, **kwargs)
        self.base_dir = os.path.dirname(os.path.realpath(sys.argv[0]))
        self.user_dir = os.path.join(self.base_dir, "user")
        self.conc_dir = os.path.join(self.user_dir, "concordances")
        self.root = root
        self.main = main
        self.draw_ui()

    def draw_ui(self):
        # init vars
        self.search_str = tk.StringVar()
        self.start_year = tk.IntVar()
        self.end_year = tk.IntVar()
        self.context_left = tk.IntVar()
        self.context_right = tk.IntVar()
        self.case_sensitive = tk.IntVar()
        self.whole_words = tk.IntVar()
        self.show_pos = tk.IntVar()
        self.show_pos_key = tk.IntVar()
        self.use_regex = tk.IntVar()
        self.corpus = tk.StringVar()
        # frame config
        self.columnconfigure(0, weight=1)
        self.rowconfigure(0, weight=1)
        self.rowconfigure(1, weight=0)
        self.rowconfigure(2, weight=0)
        self.info_label = ttk.Label(self, text="", wraplength=600)
        self.info_label.grid(column=0, row=2, sticky="new", columnspan=5)
        self.tool_frame = ttk.Frame(self, padding="5 5 5 5")
        self.tool_frame.grid(row=1, column=0, sticky="news")
        self.display_frame = ttk.Frame(self, padding="5 5 5 5")
        self.display_frame.rowconfigure(0, weight=1)
        self.display_frame.rowconfigure(1, weight=0)
        self.display_frame.columnconfigure(0, weight=1)
        self.display_frame.columnconfigure(1, weight=0)
        self.display_frame.grid(row=0, column=0, sticky="news")
        #  display frame
        self.kwic_header = ["#", "left", "key", "right", "year", "sex", "role", "class", "hisclass", 
                       "trial", "scribe", "publisher", "printer", "hiscoCode", "hiscoLabel"]
        self.kwic_tree = ttk.Treeview(self.display_frame, columns=self.kwic_header, show="headings")
        self.ybar = ttk.Scrollbar(self.display_frame, orient="vertical",
                                  command=self.kwic_tree.yview)
        self.xbar = ttk.Scrollbar(self.display_frame, orient="horizontal",
                                  command=self.kwic_tree.xview)
        self.kwic_tree.configure(yscrollcommand=self.ybar.set, xscrollcommand=self.xbar.set)
        for col in self.kwic_header:
            self.kwic_tree.heading(col, text=col.upper(),
                                   command=lambda c=col: self.sort_kwic_by(self.kwic_tree, c, 0))
        sample_left = 40*"w"
        sample_right = 40*"w"
        self.kwic_tree.column("#", width=tkFont.Font().measure("10000"))
        self.kwic_tree.column("left", width=tkFont.Font().measure(sample_left))
        self.kwic_tree.column("right", width=tkFont.Font().measure(sample_right))
        self.kwic_tree.column("year", width=tkFont.Font().measure("1913"))
        self.kwic_tree.column("sex", width=tkFont.Font().measure("SEX"))
        self.kwic_tree.column("hiscoCode", width=tkFont.Font().measure("HISCOCODE"))
        self.kwic_tree.column("role", width=tkFont.Font().measure("Interpreter"))
        self.kwic_tree.column("class", width=tkFont.Font().measure("higher (1-5)"))
        self.kwic_tree.column("hisclass", width=tkFont.Font().measure("HISCLASS"))
        self.kwic_tree.column("trial", width=tkFont.Font().measure("t19100208-1245"))
        self.kwic_tree.column("hiscoLabel", width=tkFont.Font().measure("Policeman and other Maintainers of Law and Order (except Military)"))
        self.kwic_tree.column("scribe", width=tkFont.Font().measure("James Drover Barnett , Alexander Buckler"))
        self.kwic_tree.column("publisher", width=tkFont.Font().measure("James Drover Barnett , Alexander Buckler"))
        self.kwic_tree.column("printer", width=tkFont.Font().measure("The Argus Printing Company"))
        self.ybar.grid(row=0, column=1, sticky="news")
        self.xbar.grid(row=1, column=0, sticky="news")
        self.kwic_tree.grid(row=0, column=0, sticky='news')
        #  tool frame
        tk.Label(self.tool_frame, text="Search string").grid(column=0, row=1, sticky="nw")
        #tk.Label(self.tool_frame, text="Export name").grid(column=0, row=2, sticky="nw")
        tk.Label(self.tool_frame, text="From year").grid(column=0, row=3, sticky="nw")
        tk.Label(self.tool_frame, text="to year").grid(column=2, row=3, sticky="nw")
        tk.Label(self.tool_frame, text="Context left").grid(column=0, row=4, sticky="nw")
        tk.Label(self.tool_frame, text="Context right").grid(column=2, row=4, sticky="nw")
        self.search_str_entry = ttk.Entry(self.tool_frame, width=64, textvariable=self.search_str)
        self.start_year_entry = ttk.Entry(self.tool_frame, width=4, textvariable=self.start_year)
        self.end_year_entry = ttk.Entry(self.tool_frame, width=4, textvariable=self.end_year)
        self.left_entry = ttk.Entry(self.tool_frame, width=4, textvariable=self.context_left)
        self.right_entry = ttk.Entry(self.tool_frame, width=4, textvariable=self.context_right)
        self.use_regex_check = ttk.Checkbutton(self.tool_frame, text="Regular expressions", variable=self.use_regex, onvalue=1, offvalue=0)
        self.case_sensitive_check = ttk.Checkbutton(self.tool_frame, text="Case sensitive", variable=self.case_sensitive, onvalue=1, offvalue=0)
        self.whole_words_check = ttk.Checkbutton(self.tool_frame, text="Whole words", variable=self.whole_words, onvalue=1, offvalue=0)
        #self.show_pos_key_check = ttk.Checkbutton(self.tool_frame, text="Show POS tags (key)", variable=self.show_pos_key, onvalue=1, offvalue=0)
        #self.show_pos_check = ttk.Checkbutton(self.tool_frame, text="Show POS tags (all)", variable=self.show_pos, onvalue=1, offvalue=0)
        self.obc_radio = ttk.Radiobutton(self.tool_frame, text="OBC2", variable=self.corpus, value="OBC2").grid(column=0, row=6, sticky="w")
        self.obc_pos_radio = ttk.Radiobutton(self.tool_frame, text="OBC2POS", variable=self.corpus, value="OBC2POS").grid(column=1, row=6, sticky="w")
        self.search_str_entry.grid(column=1, row=1, columnspan=6, sticky="nwe")
        #self.label_entry.grid(column=1, row=2, columnspan=6, sticky="new")
        self.start_year_entry.grid(column=1, row=3, sticky="nw")
        self.end_year_entry.grid(column=3, row=3, sticky="nw")
        self.left_entry.grid(column=1, row=4, sticky="nw")
        self.right_entry.grid(column=3, row=4, sticky="nw")
        self.use_regex_check.grid(column=0, row=5, sticky="new")
        self.whole_words_check.grid(column=1, row=5, sticky="new")
        self.case_sensitive_check.grid(column=2, row=5, sticky="new")
        #self.show_pos_key_check.grid(column=3, row=5, sticky="new")
        #self.show_pos_check.grid(column=4, row=5, sticky="new")
        # set defaults
        self.search_str.set("")
        self.start_year.set("1720")
        self.end_year.set("1913")
        self.context_left.set("40")
        self.context_right.set("40")
        self.whole_words.set(1)
        self.use_regex.set(1)
        self.case_sensitive.set(0)
        self.show_pos_key.set(0)
        self.show_pos.set(0)
        self.corpus.set("OBC2")
        # basic menu
        self.search_button = ttk.Button(self.tool_frame, text="Start search",
                                        command=self.on_search_click)
        self.search_button.grid(column=0, row=7, columnspan=3, sticky="ew")
        self.export_csv_button = ttk.Button(self.tool_frame, text="Export to .csv",
                                            command=self.on_export_csv_click)
        self.export_csv_button.grid(column=3, row=7, columnspan=1, sticky="ew")
        self.export_csv_button["state"] = "disabled"
        self.export_xlsx_button = ttk.Button(self.tool_frame, text="Export to .xlsx", command=self.on_export_xlsx_click)
        self.export_xlsx_button.grid(column=4, row=7, columnspan=1, sticky="ew")
        self.export_xlsx_button["state"] = "disabled"
        self.export_html_button = ttk.Button(self.tool_frame, text="Export to .html", command=self.on_export_html_click)
        self.export_html_button.grid(column=5, row=7, columnspan=1, sticky="ew")
        self.export_html_button["state"] = "disabled"
        self.export_tex_button = ttk.Button(self.tool_frame, text="Export to .tex", command=self.on_export_tex_click)
        self.export_tex_button.grid(column=6, row=7, columnspan=1, sticky="ew")
        self.export_tex_button["state"] = "disabled"
        
        self.bind('<Return>', self.on_search_click)
        self.search_str_entry.focus()
        for child in self.tool_frame.winfo_children():
            child.grid_configure(padx=5, pady=5)


    def strip_pos_tags(self):
        s = re.sub(r'_[^\s]+', r'', s)
        return s


    def sort_kwic_by(self, tree, col, descending):
        # grab values to sort
        if col in ["#"]:
            data = [(int(tree.set(child, col)), child) \
                for child in tree.get_children('')]
        else:
            data = [(tree.set(child, col), child) \
                for child in tree.get_children('')]
        # now sort the data in place
        data.sort(reverse=descending)
        for ix, item in enumerate(data):
            tree.move(item[1], '', ix)
        # switch the heading so it will sort in the opposite direction
        tree.heading(col, command=lambda col=col: self.sort_kwic_by(tree, col, int(not descending)))


    def on_search_click(self):
        if self.search_button["text"] == "Start search":
            self.search_cancelled = False
            self.search_button.configure(text="Cancel", state="normal")
            self.info_label.config(text="Searching. Please wait.")
            self.root.update_idletasks()
            """ evaluate search options and compile regex """
            if self.case_sensitive.get() == 1:
                self.flags = re.UNICODE
            else:
                self.flags = re.UNICODE | re.IGNORECASE
            if self.whole_words.get() == 1:
                b = r"\b"
            else:
                b = ""
            if self.use_regex.get() == 1:
                s = r'{0}{1}{0}'.format(b, self.search_str.get())
                self.regex = re.compile(s, flags=self.flags)
            else:
                s = r'{0}{1}{0}'.format(
                    b, re.escape(self.search_str.get().strip()))
                self.regex = re.compile(s, flags=self.flags)
            if len(self.search_str.get().strip()) > 0:
                self.do_search()
        else:
            self.search_button["text"] = "Start search"
            self.search_cancelled = True
            self.info_label["text"] = "Search cancelled."
            self.concordance_df = pd.DataFrame()
            self.search_job.terminate()
            self.pipe.close()
            try:
                self.root.after_cancel(self.after_job)
            except AttributeError:
                pass
            self.root.update_idletasks()
            
    def do_search(self):
        self.concordance = []
        self.clear_concordance_view()
        self.export_csv_button["state"] = "disabled"
        self.export_xlsx_button["state"] = "disabled"
        self.export_tex_button["state"] = "disabled"
        self.export_html_button["state"] = "disabled"
        start_year = self.start_year.get()
        end_year = self.end_year.get()
        options = {}
        if self.corpus.get() == "OBC2POS":
            pos = True
        else:
            pos = False
        options['pos'] = pos
        options['start_year'] = self.start_year.get()
        options['end_year'] = self.end_year.get()
        options['search_string'] = self.search_str.get()
        options['regex'] = self.regex
        options['context_left'] = self.context_left.get()
        options['context_right'] = self.context_right.get()
        params = {} # for output only  
        params['From year'] = options['start_year']
        params['To year'] = options['end_year']
        params['Search text'] = options['search_string'] 
        params['Context left'] = options['context_left']
        params['Context right'] = options['context_right']
        params['Search POS-tagged version'] = options['pos']
        params['Case sensitive search'] = self.case_sensitive.get()
        params['Match whole words'] = self.whole_words.get()
        params['Search as regular expression'] = self.use_regex.get()
        self.search_start_time = datetime.datetime.now()
        readable_time = self.search_start_time.isoformat().replace("T"," ")
        dotpos = readable_time.find(".")
        readable_time = readable_time[:dotpos]
        params['Search executed'] = readable_time
        params_list = [params]
        self.params_df = pd.DataFrame.from_dict(params_list)
        bool_cols = self.params_df.select_dtypes(include=['bool'])
        for col in bool_cols:
            self.params_df[col] = self.params_df[col].astype(int)  # False to 0, True to 1
        manager = mp.Manager()
        self.pipe, worker_pipe = mp.Pipe(duplex=False)
        self.return_shared = manager.list()
        self.search_job = mp.Process(target=obc2conc.worker_search_files,
                                     args=(options, self.return_shared, worker_pipe,))
        self.search_job.start()
        self.root.update_idletasks()
        self.root.after(1000, self.check_search_status)

    def check_search_status(self):
        try:
            if self.search_job.is_alive():
                try:
                    if self.pipe.poll(timeout=1):
                        num_lines, num_files, per = self.pipe.recv()
                        self.info_label["text"] = "Search started {}. Found >{} hits in >{} files so far (~{}% done)".format(self.search_start_time.strftime("%Y-%m-%d %H:%M:%S"), num_lines, num_files, per)
                        self.root.update_idletasks()
                    else:
                        right_now = datetime.datetime.now() 
                        diff = (right_now-self.search_start_time)/ datetime.timedelta(seconds=1)
                        self.info_label["text"] = "Searched for {} seconds so far.".format(diff)
                except BrokenPipeError as e:
                    pass
                try:
                    self.after_job = self.root.after(100, self.check_search_status)
                except AttributeError:
                    pass
            else:
                self.search_job.terminate()
                try:
                    self.root.after_cancel(self.after_job)
                except AttributeError:
                    pass
                self.search_button["text"] = "Start search"
                if self.search_cancelled:
                    self.clear_concordance_view()
                else:
                    self.concordance_df = self.return_shared[0]
                    self.finish_search()
        except AttributeError as e:
            print(e)
            try:
                self.root.after_cancel(self.after_job)
            except AttributeError:
                pass
            self.pipe.close()
            if self.search_cancelled:
                self.clear_concordance_view()
            else:
                self.concordance_df = self.return_shared[0]
                self.finish_search()
                

    def finish_search(self):
        self.search_end_time = datetime.datetime.now() 
        diff_in_min = (self.search_end_time-self.search_start_time)/ datetime.timedelta(seconds=1)
        if len(self.concordance_df.index) > 0:
            self.concordance_df = self.concordance_df.sort_values(by="event")
            num_lines = len(self.concordance_df.index)
            r = list(range(1, num_lines+1))
            random.shuffle(r)
            self.concordance_df['random'] = r
            # np.random.choice(range(1, num_lines+1), self.concordance_df.shape[0])  # add random integer from 1 to number_of_lines
            self.info_label["text"] = "{} results were found. The search took {} seconds.".format(len(self.concordance_df.index), diff_in_min)
            self.export_csv_button["state"] = "normal"
            self.export_xlsx_button["state"] = "normal"
            self.export_tex_button["state"] = "normal"
            self.export_html_button["state"] = "normal"
            self.display_concordance_view()
        else:
            self.info_label.config(text="Zero results were found. The search took {} seconds.".format(diff_in_min))

    def clear_concordance_view(self):
        self.kwic_tree.delete(*self.kwic_tree.get_children())

    def add_concordance_line(self, ind, row):
        left = row['left'].rjust(self.context_left.get())
        right = row['right'].ljust(self.context_right.get())
        key = row['key']
        row["#"] = len(self.kwic_tree.get_children()) +1 # number of rows in tree
        conc_line = "{0}\t\t{1}\t\t{2}\n".format(left, key, right)
        columns = [row[k] for k in self.kwic_header]
        self.kwic_tree.insert('', 'end', values=columns)
   
    def display_concordance_view(self):
        self.clear_concordance_view()
        update_every = 50
        if len(self.concordance_df.index) > 0:
            for ind, row in self.concordance_df.iterrows():
                self.add_concordance_line(ind, row)
                if ind % update_every == 0:
                    self.root.update()
        self.root.update_idletasks()
            
    def on_export_csv_click(self):
        col_order = ['left', 'key', 'right', 'year', 'sex', 'role', 'class', 'hisclass',
                     'trial', 'scribe', 'publisher', 'printer', 'editor', 'hiscoCode', 
                     'hiscoLabel', 'speaker', 'p2', 'p3', 'p4', 'p5', 'p6',
                     'wc', 'event', 'n', 'nTrial', 'filename', 'random']
        if len(self.concordance_df.index)> 0:
            options = {}
            options['defaultextension'] = ".csv"
            options['filetypes'] = [('CSV files', '.csv')]
            options['initialdir'] = self.conc_dir
            options['initialfile'] = 'concordance.csv'
            options['parent'] = self
            options['title'] = 'Save'
            file_str = filedialog.asksaveasfilename(**options)
            if file_str:
                self.concordance_df.to_csv(file_str, columns=col_order)
                auto = bool(self.main.conf_frame.auto_open.get())    
                if auto:
                    obc2conc.open_file_externally(file_str)
            

    def on_export_html_click(self):
        pd.set_option('display.max_colwidth', -1)
        col_order = ['left', 'key', 'right', 'year', 'sex', 'role', 'class', 'hisclass',
                     'trial', 'scribe', 'publisher', 'printer', 'editor', 'hiscoCode', 
                     'hiscoLabel', 'speaker', 'p2', 'p3', 'p4', 'p5', 'p6',
                     'wc', 'event', 'n', 'nTrial', 'filename', 'random']
        if len(self.concordance_df.index)> 0:
            options = {}
            options['defaultextension'] = ".html"
            options['filetypes'] = [('HTML files', '.html')]
            options['initialdir'] = self.conc_dir
            options['initialfile'] = 'concordance.html'
            options['parent'] = self
            options['title'] = 'Save'
            file_str = filedialog.asksaveasfilename(**options)
            if file_str:
                html = self.concordance_df.to_html(columns=col_order)
                with open(file_str, "w") as h:
                    h.write(html)
                auto = bool(self.main.conf_frame.auto_open.get())    
                if auto:
                    obc2conc.open_file_externally(file_str)

    def on_export_tex_click(self):
        col_order = ['left', 'key', 'right', 'year', 'sex', 'role', 'class', 'hisclass',
                     'trial', 'scribe', 'publisher', 'printer', 'editor', 'hiscoCode', 
                     'hiscoLabel', 'speaker', 'p2', 'p3', 'p4', 'p5', 'p6',
                     'wc', 'event', 'n', 
                     'nTrial', 'filename', 'random']
        if len(self.concordance_df.index)> 0:
            options = {}
            options['defaultextension'] = ".tex"
            options['filetypes'] = [('LaTeX files', '.tex')]
            options['initialdir'] = self.conc_dir
            options['initialfile'] = 'concordance.tex'
            options['parent'] = self
            options['title'] = 'Save'
            file_str = filedialog.asksaveasfilename(**options)
            if file_str:
                tex = self.concordance_df.to_latex(columns=col_order)
                with open(file_str, "w") as h:
                    h.write(tex)


    def on_export_xlsx_click(self):
        col_order = ['left', 'key', 'right', 'year', 'sex', 'role', 'class', 'hisclass',
                     'trial', 'scribe', 'publisher', 'printer', 'editor', 'hiscoCode', 
                     'hiscoLabel', 'speaker', 'p2', 'p3', 'p4', 'p5', 'p6',
                     'wc', 'event', 'n', 'nTrial', 'filename', 'random']
        p_col_order = ['Search text', 'From year', 'To year', 'Case sensitive search',
                       'Match whole words', 'Search POS-tagged version', 
                       'Search as regular expression','Context left', 'Context right', 'Search executed']
        if len(self.concordance_df.index > 0):

            options = {}
            options['defaultextension'] = ".xlsx"
            options['filetypes'] = [('Excel files', '.xlsx')]
            options['initialdir'] = self.conc_dir
            options['initialfile'] = 'concordance.xlsx'
            options['parent'] = self
            options['title'] = 'Save'
            file_str = filedialog.asksaveasfilename(**options)
            if file_str:
                writer = pd.ExcelWriter(file_str, engine='xlsxwriter')
                workbook = writer.book
                self.concordance_df.to_excel(writer, sheet_name="Concordance", columns=col_order)  
                self.params_df.to_excel(writer, sheet_name="Concordance parameters", columns=p_col_order, index=False)  
 
                writer.save()
                auto = bool(self.main.conf_frame.auto_open.get())    
                if auto:
                    obc2conc.open_file_externally(file_str)              
